/*
  (C) 2005-2007  Petr Lastovicka

  contents of this file are subject to the Reciprocal Public License ("RPL")
*/
#include "hdr.h"
#include "sudoku.h"

int 
 fontH[2];
HFONT hFont[2];

//---------------------------------------------------------------------------
int X(int x)
{
  return x*gridx;
}

int Y(int y)
{
  int r= y*gridy;
  if(toolBarVisible) r+=toolH;
  return r;
}

int s2x(Tsquare *s)
{
  return int(s-board)/extenty;
}

int s2y(Tsquare *s)
{
  return int(s-board)%extenty;
}

TCHAR num2char(int v)
{
  return (TCHAR)((v>=9) ? ('A'-9+v) : ((v<0) ? '.' : ('1'+v)) );
}

int char2num(TCHAR c)
{
  if(c=='.') return -1;
  if(c>='A' && c<='Z') return c-'A'+9;
  if(c>='a' && c<='z') return c-'a'+9;
  if(c>='1' && c<='9') return c-'1';
  return -2;
}
//---------------------------------------------------------------------------
bool test(Tsquare *s)
{
  int k,j,num;
  Tline *e;
  Tsquare *s1;
  Tgroup *g;

  num=s->val;
  if(num>=0 && num<Msymbol){
    for(j=0; j<s->lines; j++){
      e=s->ptr[j];
      for(k=0; k<e->len; k++){
        s1=e->ptr[k];
        if(s1->val==num && s!=s1) return true;
      }
    }
    g=s->group;
    if(g){
      for(j=0; j<g->len; j++){
        s1=g->square[j];
        if(s1->val==num && s!=s1) return true;
      }
    }
  }
  return false;
}

bool testSum(Tsquare *t)
{
  int j,S;
  Tsquare *s,*m;
  Tgroup *g;
  
  g=t->group;
  if(!g) return false;
  m=g->square[0];
  S=g->len;
  for(j=0; j<g->len; j++){
    s=g->square[j];
    if(s<m) m=s;
    if(s->val<0) return false;
    S+=s->val;
  }
  return t==m && S!=g->sum;
}

int testGreater(Tsquare *t)
{
  int j,q,result;
  Tsquare *s;

  result=0;
  if(greater && t->val>=0){
    for(j=0; j<4; j++){
      q= t->sign[j];
      if(q){
        s= t->neighbour[j];
        if(s->val>=0 && (q==1 ? s->val<=t->val : s->val>=t->val)){
          result|= 1<<j;
        }
      }
    }
  }
  return result;
}

int testConsecutive(Tsquare *t)
{
  int j,result;
  Tsquare *s;

  result=0;
  if(consecutive && t->val>=0){
    for(j=0; j<4; j++){
      s= t->neighbour[j];
      if(s && s->val>=0 && 
        (s->val==t->val+1 || s->val==t->val-1) != t->cons[j] ){
        result|= 1<<j;
      }
    }
  }
  return result;
}

bool testTotal()
{
  int i,S,S0,n;
  Tgroup *g;

  S=n=0;
  for(i=0; i<Ngroup; i++){
    g=&group[i];
    if(g->len>0){
      n+=g->len;
      S+=g->sum;
    }
  }
  S0=Nsquare*(Nsymbol+1)/2;
  if(n<Nsquare || S==S0) return false;
  msglng(800,"Total sum of all groups is %d, but it should be %d",S,S0);
  return true;
}

void checkErr()
{
  int i,eg,e,es,ec;
  Tsquare *t;

  for(i=0; i<Nboard; i++){
    t=&board[i];
    e=test(t);
    es=testSum(t);
    eg=testGreater(t);
    ec=testConsecutive(t);
    if(e!=t->err || es!=t->errSum || eg!=t->errGreater || ec!=t->errCons){
      if(showErr || t->err==2) invalidateS(t);
      t->err=e;
      t->errSum=es;
      t->errGreater=eg;
      t->errCons=ec;
    }
  }
}

bool isFinished()
{
  int i;
  Tsquare *s;

  if(done!=Nsquare) return false;
  for(i=0; i<Nboard; i++){
    s=&board[i];
    if(test(s) || testSum(s) || testGreater(s) || testConsecutive(s)) return false;
  }
  return true;
}
//---------------------------------------------------------------------------
void paintSign(HDC dc, int x, int y, int o, int err)
{
  int d,dx,dy;
  HGDIOBJ oldP=0;
  static const int VX[]={0,0,-1,1}, VY[]={-1,1,0,0};

  if(err && showErr){
    oldP=SelectObject(dc,CreatePen(PS_SOLID,0,colors[clGreaterError]));
  }
  d=max(gridy/9,2);
  dx=VX[o]*d; dy=VY[o]*d;
  MoveToEx(dc, x-dx, y-dy, 0);
  LineTo(dc, x+dx+dy, y+dy+dx);
  MoveToEx(dc, x-dx, y-dy, 0);
  LineTo(dc, x+dx-dy, y+dy-dx);
  if(err && showErr){
    DeleteObject(SelectObject(dc,oldP));
  }
}

void paintCons(HDC dc, int x, int y, int o, bool q, int err)
{
  int dx,dy,d;
  HGDIOBJ oldP=0;

  if(!showErr) err=0;
  if(!q && !err) return;
  if(err){
    oldP=SelectObject(dc,CreatePen(PS_SOLID,2,colors[clConsError]));
  }
  dx=dy=0;
  d= q ? 4 : 8;
  if(!o) dx=max(gridx/d,3);
  else dy=max(gridy/d,3);
  MoveToEx(dc, x-dx, y-dy, 0);
  LineTo(dc, x+dx, y+dy);
  if(err){
    DeleteObject(SelectObject(dc,oldP));
  }
}

void addVertex(void *ptr, int i, int dir, void *)
{
  static int gx[]={0,0,1,1};
  static int gy[]={0,1,1,0};
  static int bx[]={1,1,-1,-1};
  static int by[]={1,-1,-1,1};

  dir&=3;
  POINT *&p = *(POINT **)ptr;
  p->x=X(i/extenty+gx[dir])+groupPad*bx[dir]-(bx[dir]<0);
  p->y=Y(i%extenty+gy[dir])+groupPad*by[dir]-(by[dir]<0);
  p++;
}

void paintGroup(HDC dc, Tgroup *g)
{
  int j,k;
  Tsquare *s,*m;
  POINT pt[Mboard*2+2],*p;
  TCHAR buf[10];

  if(!g->len) return;
  for(k=0; k<g->len; k++){ 
    g->square[k]->tmp=0;
  }
  for(k=0; k<g->len; k++){ 
    if(g->square[k]->tmp) continue; //don't paint same polygon twice
    p=pt;
    borderGroup(g, g->square[k]-board, &p, &addVertex, 0);
    assert(p<=endA(pt));
    Polygon(dc,pt,p-pt);
  }
  //find top left corner
  m=g->square[0];
  for(j=0; j<g->len; j++){
    s=g->square[j];
    if(s<m) m=s;
  }
  //print sum
  SetTextColor(dc, colors[m->errSum && showErr ? clSumError : clSum]);
  SetTextAlign(dc,TA_LEFT|TA_TOP);
  SetBkMode(dc,OPAQUE);
  _itot(g->sum,buf,10);
  HGDIOBJ oldF= SelectObject(dc,GetStockObject(SYSTEM_FONT));
  TextOut(dc, X(s2x(m))+2, Y(s2y(m))+2, buf,_tcslen(buf));
  SelectObject(dc,oldF);
}

void paint(HDC dc, RECT *clip)
{
  int x,y,i,x1,y1,v,h[2],x0,y0,ng,gridx1,gridy1;
  TCHAR c;
  Tsquare *s;
  Tgroup *g;
  HGDIOBJ oldP,oldB;
  static LOGFONT font;
  RECT rc;
  static int oldDone,oldNg;
  const int kdx=15;
  HBRUSH clbr[Msymbol];
  HBRUSH brBkgnd= GetSysColorBrush(COLOR_BTNFACE);

  if(extentx==0) return;
  gridx=width/extentx;
  y0= toolBarVisible ? toolH : 0;
  GetClientRect(statusbar,&rc);
  gridy=(height-(statusBarVisible ? rc.bottom:0)-y0)/extenty;
  gridx1=gridx-2; gridy1=gridy-2;
  if(killer){ gridx1-=groupPad*2+kdx; gridy1-=groupPad*2; }
  if(greater){ gridx1-=signPad*2; gridy1-=signPad*2; }
  else if(consecutive){ gridx1-=consPad*2; gridy1-=consPad*2; }
  //create fonts
  for(i=0; i<2; i++){
    h[i]= int((i ? gridy1/sizey*0.85 : gridy*0.8));
    if(h[i]!=fontH[i]){
      fontH[i]=h[i];
      memset(&font,0,sizeof(LOGFONT));
      font.lfHeight=-h[i];
      _tcscpy(font.lfFaceName,_T("Arial"));
      DeleteObject(hFont[i]);
      hFont[i]=CreateFontIndirect(&font);
    }
  }
  //create brushes
  if(symbol==2){
    for(i=0; i<Nsymbol; i++){
      clbr[i]=CreateSolidBrush(colors[i]);
    }
  }
  SelectObject(dc,CreateSolidBrush(colors[clBkgnd]));
  SetBkColor(dc, colors[clBkgnd]);
  oldP= SelectObject(dc,CreatePen(PS_SOLID,0,colors[clGrid]));
  SetTextAlign(dc,TA_CENTER|TA_TOP);
  SetBkMode(dc,TRANSPARENT);
  i=0;
  for(x=0; x<extentx; x++){
    for(y=0; y<extenty; y++){
      s=&board[i++];
      x0=X(x); y0=Y(y); 
      x1=X(x+1); y1=Y(y+1); 
      if(clip && (x0>clip->right || y0>clip->bottom || x1<clip->left || y1<clip->top)) continue;
      oldB=0;
      c=0;
      v=s->val;
      if(v>=Msymbol){
        SetRect(&rc,x0,y0,x1,y1);
        FillRect(dc,&rc,brBkgnd);
      }else{
        if(v>=0){ 
          switch(symbol){
          default: //digits
            c= num2char(v);
            break;
          case 1: //letters
            c= (TCHAR)('A'+v);
            break;
          case 2: //colors
            if(showErr && s->err==1) c='X';
            oldB=SelectObject(dc,clbr[v]);
            break;
          }
        }
        //paint square
        Rectangle(dc,x0,y0,x1,y1);
      }
      if(oldB) SelectObject(dc,oldB);
      if(s->err==2) c='!';
      if(v<0 && symbol==2){
        SetPixel(dc, x0+gridx/2, y0+gridy/2, 0);
      }
      if(v<0){
        //draw pencil marks
        SelectObject(dc,hFont[1]);
        SetTextColor(dc, colors[clNumber]);
        v=0;
        for(y1=0; y1<sizey; y1++){
          for(x1=0; x1<sizex; x1++){
            if(s->mark[v]){
              rc.left=x0+x1*gridx1/sizex+1;
              rc.top=y0+y1*gridy1/sizey+1;
              rc.right=x0+(x1+1)*gridx1/sizex+1;
              rc.bottom=y0+(y1+1)*gridy1/sizey+1;
              if(killer) OffsetRect(&rc,groupPad+kdx,groupPad);
              if(greater) OffsetRect(&rc,signPad,signPad);
              else if(consecutive) OffsetRect(&rc,consPad,consPad);
              TCHAR c=0;
              switch(symbol){
              default: //digits
                c= num2char(v);
                break;
              case 1: //letters
                c= (TCHAR)('A'+v);
                break;
              case 2: //colors
                FillRect(dc,&rc,clbr[v]);
                break;
              }
              if(c) TextOut(dc, (rc.left+rc.right+1)>>1, 
                rc.top+(rc.bottom-rc.top-fontH[1])/2, &c,1);
            }
            v++;
          }
        }
      }
      //print symbol
      if(c){
        SetTextColor(dc, colors[(showErr && s->err || s->err==2) ? clError : clNumber]);
        SelectObject(dc,hFont[0]);
        TextOut(dc, x0+((gridx+1)>>1), y0+(gridy-fontH[0])/2, &c,1);
      }
    }
  }
  //delete brushes
  if(symbol==2){
    for(i=0; i<Nsymbol; i++){
      DeleteObject(clbr[i]);
    }
  }
  DeleteObject(SelectObject(dc,GetStockObject(NULL_BRUSH)));
  SelectObject(dc,GetStockObject(SYSTEM_FONT));
  //border
  rc.left=X(extentx);
  rc.top=Y(0);
  rc.right=width;
  rc.bottom=height;
  FillRect(dc,&rc,brBkgnd);
  rc.left=0;
  rc.top=Y(extenty);
  FillRect(dc,&rc,brBkgnd);
  //bold grid
  DeleteObject(SelectObject(dc,CreatePen(PS_SOLID,0,colors[clBoldGrid])));
  if(sizex){
    for(i=0; i<multi; i++){
      x1=coord[i].x;
      y1=coord[i].y;
      for(x=x1; x<x1+size; x+=sizex){
        for(y=y1; y<y1+size; y+=sizey){
          Rectangle(dc, X(x), Y(y), X(x+sizex), Y(y+sizey));
        }
      }
    }
  }
  //groups
  DeleteObject(SelectObject(dc,CreatePen(PS_DASH,0,colors[clGroup])));
  ng=0;
  for(i=0; i<Ngroup; i++){
    g=&group[i];
    if(g->len){
      paintGroup(dc,g);
      ng++;
    }
  }
  DeleteObject(SelectObject(dc,CreatePen(PS_SOLID,2,colors[clCons])));
  //consecutive signs
  i=0;
  for(x=0; x<extentx; x++){
    for(y=0; y<extenty; y++){
      s=&board[i++];
      paintCons(dc,(X(x)+X(x+1))>>1,Y(y), 0, s->cons[0], s->errCons&1);
      paintCons(dc,X(x),(Y(y)+Y(y+1))>>1, 1, s->cons[1], s->errCons&2);
    }
  }
  DeleteObject(SelectObject(dc,CreatePen(PS_SOLID,0,colors[clGreater])));
  //greater than signs
  i=0;
  for(x=0; x<extentx; x++){
    for(y=0; y<extenty; y++){
      s=&board[i++];
      if(s->sign[0]){
        paintSign(dc,(X(x)+X(x+1))>>1,Y(y), s->sign[0]-1, s->errGreater&1);
      }
      if(s->sign[1]){
        paintSign(dc,X(x),(Y(y)+Y(y+1))>>1, s->sign[1]+1, s->errGreater&2);
      }
    }
  }
  DeleteObject(SelectObject(dc,oldP));
  //status bar
  if(Nboard && (oldDone!=done || oldNg!=ng)){
    oldDone=done;
    oldNg=ng;
    if(!done && ng){
      status(0,_T("%d %s"),ng,lng(660,"groups"));
    }else{
      status(0,_T("%d  (%d%%)"),done,done*100/Nsquare);
    }
  }
}

void getSquareRect(Tsquare *s, RECT *rc)
{
  int x,y;

  x=s2x(s); y=s2y(s);
  rc->left=X(x);
  rc->right=X(x+1);
  rc->top=Y(y);
  rc->bottom=Y(y+1);
}

void invalidateS(Tsquare *s)
{
  RECT rc;
  getSquareRect(s,&rc);
  InvalidateRect(hWin,&rc,TRUE);
}

void invalidate()
{
  InvalidateRect(hWin,0,TRUE);
}
//---------------------------------------------------------------------------
Tsquare *hitTest(LPARAM p, int *greater)
{
  int x,y,k,d,m,rx[4],ry[4];
  Tsquare *s;
  RECT rc;

  x=LOWORD(p)/gridx;
  y=(HIWORD(p)-(toolBarVisible ? toolH : 0))/gridy;
  if(!(x>=0 && x<extentx && y>=0 && y<extenty)) return 0;
  s=&board[x*extenty+y];
  if(greater){
    getSquareRect(s,&rc);
    rx[0]=rx[2]=(rc.left+rc.right)>>1;
    rx[1]=rc.left;
    rx[3]=rc.right-1;
    ry[1]=ry[3]=(rc.top+rc.bottom)>>1;
    ry[0]=rc.top;
    ry[2]=rc.bottom-1;
    m=0x7fffffff;
    for(k=0; k<4; k++){
      d=abs(rx[k]-LOWORD(p))+abs(ry[k]-HIWORD(p));
      if(d<m){
        m=d;
        *greater=k;
      }
    }
  }
  return s;
}

void stop()
{
  KillTimer(hWin,1000);
}

void select(int n)
{
  if(n<-4 || n>=size || (n==-4 && !consecutive)) return;
  if(selectedNum>=0){
    SendMessage(toolbar,TB_CHECKBUTTON,ID_SYMBOL+selectedNum,MAKELONG(FALSE,0)); 
  }else if(selectedNum==-1){
    SendMessage(toolbar,TB_CHECKBUTTON,ID_DEL,MAKELONG(FALSE,0)); 
  }else if(selectedNum==-2){
    SendMessage(toolbar,TB_CHECKBUTTON,ID_INS,MAKELONG(FALSE,0)); 
  }else if(selectedNum==-3){
    SendMessage(toolbar,TB_CHECKBUTTON,ID_SIGN,MAKELONG(FALSE,0)); 
  }else if(selectedNum==-4){
    SendMessage(toolbar,TB_CHECKBUTTON,ID_CONS,MAKELONG(FALSE,0)); 
  }
  selectedNum=n;
  if(n>=0){
    SendMessage(toolbar,TB_CHECKBUTTON,ID_SYMBOL+n,MAKELONG(TRUE,0)); 
    status(1,_T("%c"), (symbol==1) ? (TCHAR)('A'+n) : num2char(n));
  }else if(n==-1){
    status(1,_T("Del"));
    SendMessage(toolbar,TB_CHECKBUTTON,ID_DEL,MAKELONG(TRUE,0)); 
  }else if(n==-2){
    status(1,_T("Sum"));
    SendMessage(toolbar,TB_CHECKBUTTON,ID_INS,MAKELONG(TRUE,0)); 
  }else if(n==-3){
    status(1,_T(">"));
    SendMessage(toolbar,TB_CHECKBUTTON,ID_SIGN,MAKELONG(TRUE,0)); 
  }else if(n==-4){
    status(1,_T(" |"));
    SendMessage(toolbar,TB_CHECKBUTTON,ID_CONS,MAKELONG(TRUE,0)); 
  }
}

void key(int cmd)
{
  if(cmd>='A' && cmd<='Z'){
    select(cmd-(symbol==1 ? 'A' : 'A'-9));
  }
  if(cmd>='1' && cmd<='9'){
    select(cmd-'1');
  }
  if(cmd>=VK_NUMPAD1 && cmd<=VK_NUMPAD9){
    select(cmd-VK_NUMPAD1);
  }
  if(cmd==VK_SPACE || cmd==VK_DELETE){
    select(-1);
  }
  if(cmd==VK_INSERT){
    select(-2);
  }
  if(cmd==0xBC || cmd==0xBE){ // <, >
    select(-3);
  }
  if(cmd==0xDC){ // |
    select(-4);
  }
}

#ifdef _DEBUG
DWORD getTickCount()
{
  static LARGE_INTEGER freq;
  if(!freq.QuadPart){
    QueryPerformanceFrequency(&freq);
    if(!freq.QuadPart) return GetTickCount();
  }
  LARGE_INTEGER c;
  QueryPerformanceCounter(&c);
  return (DWORD)(c.QuadPart*1000/freq.QuadPart);
}
#endif

void newGame()
{
  checkMenus();
  waitOn();
#ifdef _DEBUG
  DWORD t=getTickCount();
#endif
  gener();
#ifdef _DEBUG
  status(4,_T("%d ms"),getTickCount()-t);
#endif
  waitOff();
  noScore=false;
  Nsolution=0;
  playtime=0;
  SetTimer(hWin,1000,1000,0);
  invalidate();
}
//---------------------------------------------------------------------------
void hint(Tsquare *s)
{
  if(s->val>=0) return;
  s->err=2;
  invalidateS(s);
}

void hint()
{
  int i,j,k,p;
  Tline *e;
  Tundo *u;
  Tsquare *s;

  //oneOnSquare
  for(i=0; i<Nboard; i++){
    s=&board[i];
    if(s->n==1) hint(s);
  }
  //oneInLine
  for(j=0; j<Nline; j++){
    e=&line[j];
    for(k=0; k<Nsymbol; k++){
      if(e->number[k]==1){
        for(i=0; i<e->len; i++){
          s=e->ptr[i];
          if(!s->mask[k]) hint(s);
        }
      }
    }
  }
  for(i=0; i<Nboard; i++){
    if(board[i].err==2) return; //some hints were found
  }
  p=undoPos;
  inHint=done+1;
  resolve1();
  inHint=0;
  redoPos=undoPos;
  undoTo(p);
  if(redo()){
    u=&undoRec[undoPos-1];
    assert(u->action==0 || u->action==1);
    s=u->s;
    undo();
    hint(s);
  }
  freeGroups();
}
//---------------------------------------------------------------------------
void undoAll()
{
  if(!redoPos) redoPos=undoPos;
  while(undoPos>0){
    undo1();
    Tundo *u=&undoRec[undoPos];
    if(u->action<2) invalidateS(u->s);
    if(undoPos==undoAllPos) break;
  }
}

void showMarks()
{
  int i,j;
  Tsquare *t;

  for(i=0; i<Nboard; i++){
    t=&board[i];
    for(j=0; j<Nsymbol; j++){
      t->mark[j]= (t->mask[j]==0);
    }
  }
  invalidate();
}

void delAllMarks()
{
  int i,j;
  Tsquare *t;

  for(i=0; i<Nboard; i++){
    t=&board[i];
    for(j=0; j<Nsymbol; j++){
      t->mark[j]=false;
    }
  }
  invalidate();
}

//---------------------------------------------------------------------------
const int sizeTab[]={6,9,12,6,9,12,6,9,6,9,6,9,6,9};
const int diagTab[]={0,0,0,1,1,1,0,0,1,1,0,0,1,1};
const int killerTab[]={0,0,0,0,0,0,1,1,1,1,0,0,0,0};
const int greaterTab[]={0,0,0,0,0,0,0,0,0,0,1,1,1,1};

TscoreTab *findScoreTab(BYTE gameType, BYTE size, BYTE diag, BYTE killer, BYTE greater, BYTE cons)
{
  for(TscoreTab *t=score; t; t=t->next){
    if(t->gameType==gameType && t->size==size && t->diag==diag && 
      t->killer==killer && t->greater==greater && t->cons==cons)
      return t;
  }
  return 0;
}

TscoreTab *getScoreTab(BYTE gameType, BYTE size, BYTE diag, BYTE killer, BYTE greater, BYTE cons)
{
  TscoreTab *t= findScoreTab(gameType,size,diag,killer,greater,cons);
  if(!t){
    t=new TscoreTab;
    t->next=score;
    score=t;
    t->lastScore=-1;
    t->gameType=gameType;
    t->size=size;
    t->diag=diag;
    t->killer=killer;
    t->greater=greater;
    t->cons=cons;
    memset(t->score,0,sizeof(t->score));
  }
  return t;
}

void addScore()
{
  int i,dlg;
  TScore s;
  SYSTEMTIME t;
  
  if(playtime<=0 || playtime>65535 || symetric || level ||
    (size!=6 && size!=9 && size!=12)) return;
  readScore();
  //fill in a score record
  GetLocalTime(&t);
  s.date.wDay=(BYTE)t.wDay;
  s.date.wMonth=(BYTE)t.wMonth;
  s.date.wYear=t.wYear;
  s.date.wMinute=(BYTE)t.wMinute;
  s.date.wHour=(BYTE)t.wHour;
  s.playtime=(WORD)playtime;
  //get table
  TscoreTab *a= getScoreTab((BYTE)gameType,(BYTE)size,(BYTE)diag,(BYTE)killer,(BYTE)greater,(BYTE)consecutive);
  TScore *score= a->score;
  //InsertSort
  for(i=0; i<Dscore; i++){
    if( s.playtime < score[i].playtime || !score[i].playtime) break;
  }
  if(i<Dscore){
    //ask for player name
    dlg= (int) DialogBoxParam(inst, MAKEINTRESOURCE(IDD_NAME),
      hWin, (DLGPROC) NameProc, 0);
    convertT2W(playerName,w);
    wcscpy(s.name,w);
    if(dlg!=IDCANCEL){
      //add new record to the score table
      a->lastScore=i;
      for(int j=Dscore-1; j>i; j--)  score[j]=score[j-1];
      score[i]=s;
      //save to disk
      writeScore();
      //show table
      DialogBoxParam(inst, MAKEINTRESOURCE(IDD_HISCORE), hWin, (DLGPROC) ScoreProc, 0);
    }else{
      //delete empty table
      if(!score[0].playtime){
        assert(a==::score);
        ::score=a->next;
        delete a;
      }
    }
  }
}
//---------------------------------------------------------------------------
void parseScore(BYTE *data, int len)
{
  int i,n,Nname;
  WCHAR *w, **nameA=0;
  BYTE b;
  WORD d;
  TscoreTab *t;
  TScore *r;
  BYTE *p=data;

  if(!strncmp((char*)data,"Color Sudoku score 2",20)){
    try
    {
      p+=20;
      //names
      Nname= *(WORD*)p;
      p+=2;
      nameA= new WCHAR*[Nname];
      w=(WCHAR*)p;
      for(i=0; i<Nname; i++){
        nameA[i]=w;
        w=wcschr(w,0)+1;
      }
      p=(BYTE*)w;
      //score tables
      for(;p<data+len;){
        b=p[2];
        t=getScoreTab(p[0], p[1], (BYTE)(b&1), (BYTE)((b>>1)&1), (BYTE)((b>>2)&1), (BYTE)((b>>3)&1));
        memset(t->score,0,sizeof(t->score));
        p+=3;
        n=*p++;
        for(i=0; i<n; i++){
          r=&t->score[i];
          if(r<endA(t->score)){
            //name
            r->name[0]=0;
            d= *(WORD*)p;
            if(d<Nname) wcsncpy(r->name,nameA[d],sizeA(r->name));
            r->name[sizeA(r->name)-1]=0;
            //time and date
            r->playtime= *(WORD*)(p+2);
            memcpy(&r->date,p+4,6);
            //if(r->date.wYear<1990 || r->date.wYear>3000) r->playtime=0;
          }
          p+=10;
        }
      }
    }
    catch(...)
    {
      msg("Score table is damaged");
    }
    delete[] nameA;
  }else if(len%(Dscore*sizeof(TScore))==0){
    //old version, fixed size file
    for(i=0; i<Dtab; i++){
      if(p==data+len) break; //file is too short
      t=getScoreTab(0,(BYTE)sizeTab[i],(BYTE)diagTab[i],(BYTE)killerTab[i],(BYTE)greaterTab[i],0);
      //copy score table
      memcpy(t->score,p,Dscore*sizeof(TScore));
      p+=Dscore*sizeof(TScore);
    }
  }
}

void writeScore(BYTE *&data, int &len)
{
  TscoreTab *t;
  TScore *r;
  WCHAR **nameA;
  BYTE b,*p,*p1;
  int i,n,Nname;

  //estimate data length
  n=22;
  Nname=0;
  for(t=score; t; t=t->next){
    n+= Dscore*(sizeA(t->score->name)+10) + 4;
    Nname+= Dscore;
  }
  nameA= new WCHAR*[Nname];
  //write header
  p= data= new BYTE[n];
  strcpy((char*)p,"Color Sudoku score 2");
  p+=22;
  //write player names
  Nname=0;
  for(t=score; t; t=t->next){
    for(r=t->score; r<endA(t->score); r++){
      for(i=0; ; i++){
        if(i==Nname){
          //write new name
          wcscpy(nameA[Nname++]= (WCHAR*)p, r->name);
          p+= 2*(wcslen(r->name)+1);
          break;
        }
        //compare with previous names
        if(!wcscmp(nameA[i],r->name)) break;
      }
    }
  }
  assert(Nname<0x10000);
  *(WORD*)(data+20)= (WORD)Nname;

  //write tables
  for(t=score; t; t=t->next){
    *p++= t->gameType;
    *p++= t->size;
    *p++= (BYTE)((t->diag) | (t->killer<<1) | (t->greater<<2) | (t->cons<<3));
    p1=p;
    p++;
    for(b=0, r=t->score;  r<endA(t->score);  b++, r++){
      if(!r->playtime) break;
      for(i=0; ;i++){
        assert(i<Nname);
        if(!wcscmp(nameA[i],r->name)) break;
      }
      *(WORD*)p= (WORD)i;
      *(WORD*)(p+2)= r->playtime;
      memcpy(p+4,&r->date,6);
      p+=10;
    }
    *p1=b;
  }
  len= int(p-data);
  delete[] nameA;
}

//hiscore file name
void getHiscoreFile(TCHAR *buf)
{
  HKEY key;
  DWORD d;

  getExeDir(buf,_T(""));
  if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, _T("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders"), 
    0, KEY_QUERY_VALUE, &key)==ERROR_SUCCESS){
    d=2*(MAX_PATH-32);
    if(RegQueryValueEx(key,_T("Common AppData"),0,0, (BYTE *)buf, &d)==ERROR_SUCCESS){
      _tcscat(buf,_T("\\Petr Lastovicka\\"));
    }
    RegCloseKey(key);
  }
  _tcscat(buf,hiscoreFile);
}

void writeScore()
{
  HKEY key;
  FILE *f;
  TCHAR *t;
  BYTE *data=0;
  int len;
  TCHAR *fn=(TCHAR*)_alloca(2*MAX_PATH);
  getHiscoreFile(fn);

  if(score){
    writeScore(data,len);
    if(data && len){
      t=cutPath(fn)-1;
      *t=0;
      CreateDirectory(fn,0);
      *t='\\';
      if((f=_tfopen(fn,_T("wb")))!=0){
        //write to file
        fwrite(data,len,1,f);
        if(!fclose(f)){
          deleteini(HKEY_LOCAL_MACHINE);
        }
      }else if(RegCreateKey(HKEY_LOCAL_MACHINE, subkey, &key)==ERROR_SUCCESS){
        //write to registry
        RegSetValueEx(key,_T("BEST"),0,REG_BINARY, data, len);
        RegCloseKey(key);
      }
    }
  }else{
    //table is empty => delete it
    DeleteFile(fn);
    deleteini(HKEY_LOCAL_MACHINE);
  }
}

void readScore()
{
  HKEY key;
  FILE *f;
  int len=0;
  BYTE *data=0;
  TCHAR *fn=(TCHAR*)_alloca(2*MAX_PATH);
  getHiscoreFile(fn);
  
  //read file
  if((f=_tfopen(fn,_T("rb")))!=0){
    fseek(f,0,SEEK_END);
    len=ftell(f);
    rewind(f);
    if(len>0 && len<10000000){
      data=new BYTE[len];
      if(fread(data,len,1,f)>0){
        parseScore(data,len);
      }
    }
    fclose(f);
  }else if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, subkey, 0, KEY_QUERY_VALUE, &key)==ERROR_SUCCESS){
    //read registry
    RegQueryValueEx(key,_T("BEST"),0,0,0,(DWORD*)&len);
    if(len>0){
      data=new BYTE[len];
      RegQueryValueEx(key,_T("BEST"),0,0,data, (DWORD*)&len);
      parseScore(data,len);
    }
    RegCloseKey(key);
  }
  delete[] data;
}
//------------------------------------------------------------------
void open(TCHAR *fn)
{
  int i,x,y,v,err,n,pass,size1=0,extentx1,extenty1,diag1=0,undoPos1,gameType1,multi1,sizex1,sizey1;
  char *p;
  char buf[(Msymbol+1)*Mboard+8];
  Tsquare *grp[Mboard],*s;

  FILE *f= _tfopen(fn,_T("rt"));
  if(!f){
    msglng(730,"Cannot open file %s",fn);
  }else{
    err=0;
    //1.pass - check format, 2.pass - read data
    for(pass=0; pass<2 && !err; pass++){
      rewind(f);
      //header
      fgets(buf,sizeof(buf),f);
      p=strchr(buf,0)-1;
      if(strncmp(buf,"Sudoku",6) || *p!='\n'){ err=1; break; }
      *p=0;
      gameType1=0;
      if(buf[6]){
        //size
        size1= strtol(buf+6,&p,10);
        if(size1<=0 || *p!=' '){ err=1; break; }
        p++;
        //gameType
        gameType1=-1;
        for(i=0; gameTypeA[i].name; i++){
          if(!strcmp(gameTypeA[i].name,p)) gameType1=i;
        }
        if(gameType1<0){ err=12; break; }
      }
      //extent
      fgets(buf,sizeof(buf),f);
      i = strlen(buf)-1;
      if(gameType1==0) size1 = i;
      getExtent(gameType1,size1,extentx1,extenty1,multi1,sizex1,sizey1);
      if(size1<4 || size1>Msize || i!=extentx1){ err=2; break; }

      if(pass){
        stop();
        size=size1;
        diag=diag1;
        gameType=gameType1;
        playtime=killer=greater=consecutive=symetric=level=0;
        initSquare();
        noScore=true;
        Nsolution=0;
      }
      //board
      for(y=0; y<extenty1; y++){
        for(x=0; x<extentx1; x++){
          s=&board[x*extenty+y];
          if(buf[x]==' '){
            if(pass && s->val<Msymbol) err=3;
          }else{
            if(buf[x]==0){ err=3; goto le; }
            v=char2num(buf[x]);
            if(v<-1 || v>=size1) err=3;
            else if(pass){
              if(s->val>=Msymbol) err=3;
              else put(v,s);
            }
          }
        }
        fgets(buf,sizeof(buf)-4,f);
      }
      undoPos1=0;
      while(!feof(f)){
        if(!strncmp(buf,"Groups:",7)){
          for(p=buf+7; *p!='\n'; ){
            i=strtol(p,&p,10);
            if(i<=0 || i>9999 || *p!='{'){ err=7; break; }
            n=0;
            do{
              p++;
              x=char2num(*p++);
              y=char2num(*p++);
              if(x<0 || x>=extentx1 || y<0 || y>=extenty1 || n>=multi1*size1*size1+Mline){ err=7; break; }
              grp[n++]= &board[x*extenty1+y];
            }while(*p==',');
            if(*p!='}'){ err=7; break; }
            p++;
            if(pass){
              killer=1;
              addGroup(i,n,grp);
            }
          }
        }
        if(!strncmp(buf,"GreaterThan:",12)){
          for(p=buf+12, i=0;  *p!='\n';  p++, i++){
            v=*p-'0';
            if(v<0 || v>8 || i>=extentx1*extenty1){ err=11; break; }
            if(pass){
              s=&board[i];
              if(s->val>=Msymbol) p--;
              else{
                greater=1;
                putSign(v/3,s,0);
                putSign(v%3,s,1);
              }
            }
          }
        }
        if(!strncmp(buf,"Consecutive:",12)){
          for(p=buf+12, i=0;  *p!='\n';  p++, i++){
            v=*p-'0';
            if(v<0 || v>3 || i>=extentx1*extenty1){ err=13; break; }
            if(pass){
              s=&board[i];
              if(s->val>=Msymbol) p--;
              else{
                consecutive=1;
                if(v&2) putCons(true,s,0);
                if(v&1) putCons(true,s,1);
              }
            }
          }
        }
        if(!strncmp(buf,"Mark:",5)){
          x=y=0;
          for(p=buf+5; *p!='\n'; ){
            v=char2num(*p++);
            if(v<-1 || v>=size1 || y>=extenty1){ err=10; break; }
            if(v<0){
              x++;
              if(x==extentx1){ x=0; y++; }
            }else if(pass){
              s=&board[x*extenty+y];
              if(s->val>=Msymbol) err=10;
              else s->mark[v]=true;
            }
          }
        }
        if(!strncmp(buf,"Undo:",5)){
          for(p=buf+5; *p!='\n'; ){
            x=char2num(*p++);
            y=char2num(*p++);
            v=char2num(*p++);
            if(x<0 || y<0 || v<-1 || undoPos1>=Mundo ||
              x>=extentx1 || y>=extenty1 || v>=size1){ err=6; break; }
            if(pass){
              Tundo *u= &undoRec[undoPos1];
              u->action=0;
              u->s= &board[x*extenty+y];
              u->prev=(char)v;
            }
            undoPos1++;
          }
        }
        if(!strncmp(buf,"Diagonal:",9)){
          if(!strncmp(buf+9,"yes",3)) diag1=1;
          else if(!strncmp(buf+9,"no",2)) diag1=0;
          else err=5;
        }
        if(!strncmp(buf,"Symmetrical:",12)){
          if(!strncmp(buf+12,"yes",3)){ if(pass) symetric=1; }
          else if(!strncmp(buf+12,"no",2)){ if(pass) symetric=0; }
          else err=8;
        }
        if(!strncmp(buf,"Level:",6)){
          i=sscanf(buf+6,"%d",&x);
          if(!i || unsigned(x)>100){
            err=9;
          }else if(pass){
            level=x;
          }
        }
        if(!strncmp(buf,"Time:",5)){
          int h,m,s;
          i=sscanf(buf+5,"%d:%d:%d",&h,&m,&s);
          if(i<=1){
            err=4;
          }else if(pass){
            playtime=h*60+m;
            if(i==3) playtime=playtime*60+s;
            noScore=false;
          }
        }
        fgets(buf,sizeof(buf)-4,f);
      }
      if(pass) undoPos=undoPos1;
    }
le:
    fclose(f);
    if(err){
      msglng(731,"Invalid format of %s",fn);
    }else{
      checkMenus();
      numButtons();
      invalidate();
      SetTimer(hWin,1000,1000,0);
    }
  }
}

void save(TCHAR *fn)
{
  int i,j,x,y;
  char c;
  bool marks;
  Tsquare *s;

  FILE *f= _tfopen(fn,_T("wt"));
  if(!f){
    msglng(733,"Cannot create file %s",fn);
  }else{
    fputs("Sudoku",f);
    if(gameType>0){
      fprintf(f," %d %s", size, gameTypeA[gameType].name);
    }
    fputc('\n',f);
    marks=false;
    for(y=0; y<extenty; y++){
      for(x=0; x<extentx; x++){
        s=&board[x*extenty+y];
        if(s->val>=Msymbol) 
          fputc(' ',f);
        else{
          fputc(num2char(s->val), f);
          for(i=0; i<Nsymbol; i++) if(s->mark[i]) marks=true;
        }
      }
      fputc('\n',f);
    }
    fprintf(f,"Diagonal:%s\nSymmetrical:%s\n",
      diag ? "yes":"no", symetric ? "yes":"no");
    if(level) fprintf(f,"Level:%d\n",level);
    if(Ngroup){
      fputs("Groups:",f);
      for(i=0; i<Ngroup; i++){
        Tgroup *g= &group[i];
        if(g->len){
          fprintf(f,"%d",g->sum);
          c='{';
          for(j=0; j<g->len; j++){
            fprintf(f,"%c%c%c",c,
              num2char(s2x(g->square[j])), num2char(s2y(g->square[j])));
            c=',';
          }
          fputc('}',f);
        }
      }
      fputc('\n',f);
    }
    if(Nsign>0){
      fputs("GreaterThan:",f);
      for(i=0; i<Nboard; i++){
        s=&board[i];
        if(s->val<Msymbol)
          fputc(s->sign[0]*3 + s->sign[1] + '0', f);
      }
      fputc('\n',f);
    }
    if(consecutive>0){  //Ncons can be 0
      fputs("Consecutive:",f);
      for(i=0; i<Nboard; i++){
        s=&board[i];
        if(s->val<Msymbol)
          fputc(int(s->cons[0])*2 + int(s->cons[1]) + '0', f);
      }
      fputc('\n',f);
    }
    if(undoPos>0){
      fputs("Undo:",f);
      for(i=0; i<undoPos; i++){
        Tundo *u= &undoRec[i];
        if(u->action<2) fprintf(f,"%c%c%c", num2char(s2x(u->s)),
          num2char(s2y(u->s)), num2char(u->prev));
      }
      fputc('\n',f);
    }
    if(marks){
      fputs("Mark:",f);
      for(y=0; y<extenty; y++){
        for(x=0; x<extentx; x++){
          s=&board[x*extenty+y];
          if(s->val<Msymbol){
            for(i=0; i<Nsymbol; i++){
              if(s->mark[i]) fputc(num2char(i),f);
            }
          }
          fputc('.',f);
        }
      }
      fputc('\n',f);
    }
    if(!noScore){
      TCHAR buf[12];
      printTime(buf,playtime);
      _ftprintf(f,_T("Time:%s\n"),buf);
    }
    fclose(f);
  }
}
//---------------------------------------------------------------------------
void addGroup(int sum, int len, Tsquare **squares)
{
  int i;
  Tgroup *g;
  Tsquare *s;

  if(sum<=0 || len>Nsymbol) return;
  //find empty slot
  for(i=0; ;i++){
    if(i==Ngroup){
      if(i==Mgroup) return; 
      Ngroup++;
      break;
    }
    if(group[i].len==0) break;
  }
  //create group
  g=&group[i];
  g->sum=sum;
  g->len=0;
  g->n=0;
  for(i=0; i<len; i++){
    s=squares[i];
    if(!s->group){
      g->square[g->len++]=s;
      s->group=g;
    }
  }
}

void delGroup(Tgroup *g)
{
  if(!g) return;
  for(int i=0; i<g->len; i++){
    g->square[i]->group=0;
  }
  g->len=0;
}

void invalidateGroup(Tgroup *g)
{
  if(!g) return;
  for(int i=0; i<g->len; i++){
    invalidateS(g->square[i]);
  }
}

void insertGroup()
{
  int i,sum;
  bool z;
  Tgroup *g,*g1,*g2;
  Tsquare *s;
  
  g1=g2=0;
  z=false;
  for(i=0; i<insLen; i++){
    g=insSquares[i]->group;
    if(!g) z=true;
    else if(g!=g1 && g!=g2){
      if(!g1) g1=g;
      else if(!g2) g2=g;
      else return;
    }
  }
  if(g2){
    if(z) return;
    //join two groups
    for(i=0; i<g2->len; i++){
      (g1->square[g1->len++]=g2->square[i])->group=g1;
    }
    g1->sum+=g2->sum;
    invalidateGroup(g2);
    g2->len=0;
  }else if(!g1 || insLen==1){
    sum= DialogBoxParam(inst, MAKEINTRESOURCE(IDD_SUM), hWin, (DLGPROC)SumProc, g1 ? g1->sum : 0);
    if(sum>9999) return;
    if(sum==0){
      //delete group
      invalidateGroup(g1);
      delGroup(g1);
    }else if(!g1){
      //create new group
      addGroup(sum,insLen,insSquares);
      g1=insSquares[0]->group;
    }else{
      //change sum
      g1->sum=sum;
    }
  }else{
    //enlarge group
    for(i=0; i<insLen; i++){
      s=insSquares[i];
      if(!s->group && g1->len<Nsymbol){
        (g1->square[g1->len++]=s)->group=g1;
      }
    }
  }
  invalidateGroup(g1);
}

void insertSign(LPARAM lParam, int what)
{
  int k;
  int q;
  Tsquare *s;

  s= hitTest(lParam,&k);
  if(s && s->neighbour[k]){
    if(k>1) what=BYTE(3-what);
    if(s->sign[k]==what) q=0;
    else q=what;
    putSign(q,s,k);
    invalidateS(s);
    invalidateS(s->neighbour[k]);
    noScore=true;
    Nsolution=0;
    checkErr();
  }
}

void insertCons(LPARAM lParam)
{
  int k;
  Tsquare *s;

  s= hitTest(lParam,&k);
  if(s && s->neighbour[k]){
    putCons(!s->cons[k],s,k);
    invalidateS(s);
    invalidateS(s->neighbour[k]);
    noScore=true;
    Nsolution=0;
    checkErr();
  }
}
//------------------------------------------------------------------
void lbutton(LPARAM lParam)
{
  Tsquare *s;
  
  s= hitTest(lParam);
  if(s){
    if(selectedNum==-2){
      inserting=true;
      SetCapture(hWin);
      insSquares[0]=s;
      insLen=1;
    }else if(selectedNum==-3){
      //insert greater than sign
      insertSign(lParam,1);
    }else if(selectedNum==-4){
      //insert consecutive sign
      insertCons(lParam);
    }else if(!s->lock && s->val!=selectedNum){
      blocked++;
      put(selectedNum,s);
      blocked--;
      invalidateS(s);
      checkErr();
      if(isFinished()){
        stop();
        status(4,lng(661,"GAME OVER"));
        if(!noScore) addScore();
      }
    }
  }
}

void rbutton(LPARAM lParam)
{
  Tsquare *s= hitTest(lParam);
  if(selectedNum==-3){
    //put greater than sign
    insertSign(lParam,2);
  }else if(s && s->val<0 && selectedNum>=-1){
    if(selectedNum>=0){
      s->mark[selectedNum]= !s->mark[selectedNum]; 
    }else{
      for(int i=0; i<Nsymbol; i++){
        s->mark[i]=false;
      }
    }
    invalidateS(s);
  }
}

#ifdef _DEBUGM
void mousemove(LPARAM lParam)
{
  int i;
  TCHAR buf[Msymbol*5];

  Tsquare *s= hitTest(lParam);
  if(s){
    buf[0]=0;
    for(i=0; i<Nsymbol; i++){
      _stprintf(_tcschr(buf,0),_T(" %d,"),s->mask[i]);
    }
    _tcschr(buf,0)[-1]=0;
    status(4,buf);
  }
}
#endif
//------------------------------------------------------------------
